use gtk::prelude::*;
use gtk::{Application, ApplicationWindow, Button, FileDialog, Picture, Scale, Box as GtkBox, ComboBoxText};
use image::{DynamicImage, ImageBuffer, RgbImage, Rgb, imageops};
use rayon::prelude::*;
use std::path::{Path, PathBuf};
use std::rc::Rc;
use std::cell::RefCell;

mod filters;

fn main() {
    let app = Application::builder()
        .application_id("ru.risdeveau.imagefilters")
        .build();

    app.connect_activate(|app| {
        build_ui(app);
    });

    app.run();
}

fn build_ui(app: &Application) {
    let window = ApplicationWindow::builder()
        .application(app)
        .title("Image Filters")
        .default_width(800)
        .default_height(600)
        .build();
    window.set_resizable(false);

    let container = GtkBox::new(gtk::Orientation::Vertical, 5);

    let open_button = Button::with_label("Open Image");
    let picture = Picture::new();
    let brightness_slider = Scale::with_range(gtk::Orientation::Horizontal, -100.0, 100.0, 1.0);
    brightness_slider.set_value(0.0);

    let filter_combo = ComboBoxText::new();
    filter_combo.append_text("Original");
    filter_combo.append_text("Gray World");
    filter_combo.append_text("Histogram Stretch");
    filter_combo.append_text("Emboss");
    filter_combo.append_text("Blur");
    filter_combo.append_text("Negative");
    filter_combo.append_text("Heatmap");
    filter_combo.append_text("Pixelate");
    filter_combo.append_text("Retro");
    filter_combo.append_text("Oil");
    filter_combo.append_text("Water");
    filter_combo.append_text("Sharpen");
    filter_combo.append_text("Sepia");
    filter_combo.append_text("Shift");
    filter_combo.set_active(Some(0));

    let image_data = Rc::new(RefCell::new(None::<(PathBuf, RgbImage)>));

    fn parallel_brightness(image: &mut RgbImage, value: i16) {
        let value = value.clamp(-255, 255); // Ограничиваем диапазон

        for pixel in image.pixels_mut() {
            for channel in 0..3 {
                let new_val = pixel[channel] as i32 + value as i32;
                pixel[channel] = new_val.clamp(0, 255) as u8;
            }
        }
    }

    fn apply_filter(image: &mut RgbImage, filter_name: &str) {
        match filter_name {
            "Gray World" => filters::gray_world(image),
            "Histogram Stretch" => filters::histogram_stretch(image),
            "Emboss" => filters::emboss(image),
            "Blur" => filters::blur(image),
            "Negative" => filters::invert(image),
            "Heatmap" => filters::heatmap(image),
            "Pixelate" => filters::pixelate(image, 8),
            "Retro" => filters::retro_effect(image),
            "Oil" => filters::oil_painting(image),
            "Water" => filters::watercolor(image),
            "Sharpen" => filters::sharpen(image),
            "Sepia" => filters::sepia(image),
            "Shift" => filters::shift(image, 10, 25),
            _ => {}
        }
    }

    {
        let window = window.clone();
        let image_data = image_data.clone();
        let picture = picture.clone();
        open_button.connect_clicked(move |_| {
            let dialog = FileDialog::builder()
                .title("Select Image")
                .build();

            let filter = gtk::FileFilter::new();
            filter.add_mime_type("image/*");
            dialog.set_default_filter(Some(&filter));

            let window = window.clone();
            let image_data = image_data.clone();
            let picture = picture.clone();
            dialog.open(
                Some(&window),
                None::<&gtk::gio::Cancellable>,
                move |result| {
                    if let Ok(file) = result {
                        if let Some(path) = file.path() {
                            match image::open(&path) {
                                Ok(img) => {
                                    let mut rgb_img = img.to_rgb8();
                                    *image_data.borrow_mut() = Some((path.to_path_buf(), rgb_img.clone()));

                                    let temp_path = Path::new("/tmp/current_image.png");
                                    if rgb_img.save(&temp_path).is_ok() {
                                        picture.set_filename(Some(temp_path));
                                    }
                                }
                                Err(e) => eprintln!("Image load error: {}", e),
                            }
                        }
                    }
                }
            );
        });
    }

    {
        let image_data = image_data.clone();
        let picture = picture.clone();
        let filter_combo = filter_combo.clone();
        brightness_slider.connect_value_changed(move |slider| {
            if let Some((path, ref mut original_img)) = image_data.borrow_mut().as_mut() {
                let mut edited_img = original_img.clone();
                let brightness_value = (slider.value() * 2.55) as i16;

                parallel_brightness(&mut edited_img, brightness_value);

                if let Some(active_filter) = filter_combo.active_text() {
                    apply_filter(&mut edited_img, &active_filter);
                }

                let temp_path = Path::new("/tmp/current_image.png");
                if edited_img.save(&temp_path).is_ok() {
                    picture.set_filename(Some(temp_path));
                }
            }
        });
    }

    {
        let image_data = image_data.clone();
        let picture = picture.clone();
        let brightness_slider = brightness_slider.clone();
        filter_combo.connect_changed(move |combo| {
            if let Some((path, ref mut original_img)) = image_data.borrow_mut().as_mut() {
                let mut edited_img = original_img.clone();
                let brightness_value = (brightness_slider.value() * 2.55) as i16;

                parallel_brightness(&mut edited_img, brightness_value);

                if let Some(active_filter) = combo.active_text() {
                    apply_filter(&mut edited_img, &active_filter);
                }

                let temp_path = Path::new("/tmp/current_image.png");
                if edited_img.save(&temp_path).is_ok() {
                    picture.set_filename(Some(temp_path));
                }
            }
        });
    }

    container.append(&open_button);
    container.append(&picture);
    container.append(&brightness_slider);
    container.append(&filter_combo);
    window.set_child(Some(&container));
    window.show();
}
